import sentaku
from .ux import TodoUX
from .pseudorpc import PseudoRpc


class ViaAPI(sentaku.ApplicationImplementation):
    """access to the core api of the application"""


class ViaUX(sentaku.ApplicationImplementation):
    """access to the application via the basic api of the pseudo-ux"""

    @classmethod
    def from_api(cls, api):
        """creates a ux for the given api before

        returning the implementation holder"""
        return cls(TodoUX(api))


class ViaRPC(sentaku.ApplicationImplementation):
    """access the application via the pseudo rpc"""

    @classmethod
    def from_backend(cls, backend):
        """creates a rpc for the given backend before
        """
        return cls(PseudoRpc(backend))


class TodoItem(sentaku.Element):
    """describing a todo list element"""
    def __init__(self, parent, name):
        super(TodoItem, self).__init__(parent=parent)
        self.name = name

    @property
    def completed(self):
        raise NotImplementedError

    set_completion_state = sentaku.ImplementationRegistry()

    @set_completion_state.implemented_for(ViaAPI, ViaUX)
    def set_completion_state(self, value):
        col = self.impl.get_by(self.parent.name)
        elem = col.get_by(self.name)
        elem.completed = value

    @set_completion_state.implemented_for(ViaRPC)
    def set_completion_state(self, value):
        if value:
            self.impl.complete_item(self.parent.name, self.name)
        else:
            raise NotImplementedError('rpc cant undo completion')

    @completed.setter
    def completed(self, value):
        self.set_completion_state(value)


class TodoCollection(sentaku.Collection):
    """domain object describing a todo list"""
    def __init__(self, parent, name):
        super(TodoCollection, self).__init__(parent=parent)
        self.name = name

    create_item = sentaku.ImplementationRegistry()

    @create_item.implemented_for(ViaAPI, ViaUX)
    def create_item(self, name):
        collection = self.impl.get_by(self.name)
        elem = collection.create_item(name=name)
        assert elem
        return TodoItem(self, name=name)

    @create_item.implemented_for(ViaRPC)
    def create_item(self, name):
        self.impl.make_item(self.name, name)
        return TodoItem(self, name=name)

    get_by = sentaku.ImplementationRegistry()

    @get_by.implemented_for(ViaAPI, ViaUX)
    def get_by(self, name):
        collection = self.impl.get_by(self.name)
        elem = collection.get_by(name)
        if elem is not None:
            return TodoItem(self, name=name)

    @get_by.implemented_for(ViaRPC)
    def get_by(self, name):
        if self.impl.has_item(self.name, name):
            return TodoItem(self, name=name)

    clear_completed = sentaku.ImplementationRegistry()

    @clear_completed.implemented_for(ViaAPI, ViaUX)
    def clear_completed(self):
        collection = self.impl.get_by(self.name)
        collection.clear_completed()

    @clear_completed.implemented_for(ViaRPC)
    def clear_completed(self):
        self.impl.clear_completed(self.name)


class TodoApi(sentaku.ApplicationDescription):
    """example description for a simple todo application"""

    @classmethod
    def from_api(cls, api):
        """
        create an application description for the todo app,
        that based on the api can use either tha api or the ux for interaction
        """
        via_api = ViaAPI(api)
        via_ux = ViaUX.from_api(api)
        via_rpc = ViaRPC.from_backend(api)

        return cls.from_implementations([via_api, via_ux, via_rpc])

    create_collection = sentaku.ImplementationRegistry()

    @create_collection.implemented_for(ViaUX, ViaAPI)
    def create_collection(self, name):
        self.impl.create_item(name)
        return TodoCollection(self, name=name)

    @create_collection.implemented_for(ViaRPC)
    def create_collection(self, name):
        self.impl.make_collection(name)
        return TodoCollection(self, name=name)



import contextlib


class ImplementationContext(object):
    """ maintains a mapping
    of :ref:`implementation-identification` to implementations,
    as well as the list of currently availiable Implementations
    in the order of precedence.

    :param dict implementations:
        the implementations availiable in the context

        a mapping of :ref:`implementation-identification` to implementation
    :param default_choices:
        the implementations that should be used by default
        in order of percedence
    :type default_choices: list or None
    """
    def __init__(self, implementations, default_choices=None):
        self._implementations = implementations
        from .chooser import ChooserStack
        self.implementation_chooser = ChooserStack(default_choices)

    @property
    def impl(self):
        """the currently active implementation"""
        return self.implementation_chooser.choose(
            self._implementations).value

    @classmethod
    def from_implementations(cls, implementations, default_choices=None):
        """utility to create the application description
        by passing instances of the different implementations"""
        implementations = {
            type(implementation): implementation.implementation
            for implementation in implementations
        }
        return cls(implementations=implementations,
                   default_choices=default_choices)

    @property
    def context(self):
        """alias for consistence with elements"""
        return self

    root = context

    @contextlib.contextmanager
    def use(self, *implementation_types, **kw):
        """contextmanager for controlling
        the currently active/usable implementations and
        their order of percedence

        :param `implementation-identification` implementation_types:
            the implementations availiable within the context
        :keyword bool frozen: if True prevent further nesting
        """
        if kw:
            assert len(kw) == 1
            assert 'frozen' in kw
        with self.implementation_chooser.pushed(
                implementation_types, frozen=kw.get('frozen', False)):
            yield self.impl


class _ImplementationBindingMethod(object):
    """bound method equivalent for :class:`ImplementationCooser`

    on call it:

    * looks up the implementation
    * freezes the context
    * calls the actual implementation
    """
    def __init__(self, instance, selector):
        self.instance = instance
        self.selector = selector

    def __call__(self, *k, **kw):
        ctx = self.instance.context
        choice, implementation = ctx.implementation_chooser.choose(
            self.selector.implementations)

        bound_method = implementation.__get__(
            self.instance, type(self.instance))
        with ctx.use(choice, frozen=True):
            return bound_method(*k, **kw)


class ContextualMethod(object):
    """
    descriptor for implementing context sensitive methods
    and registration of their implementations


    .. code:: python

        class Example(Element):
            action = ContextualMethod()
            @action.implemented_for("db")
            def action(self):
                pass

           @action.implemented_for("test")
           def action(self):
               pass
    """

    def __init__(self):
        self.implementations = {}

    def __repr__(self):
        return '<ContextualMethod {implementations}>'.format(
            implementations=sorted(self.implementations.keys()))

    def _add_implementations(self, keys, func):
        for key in keys:
            assert key not in self.implementations
            self.implementations[key] = func

    def implemented_for(self, *implementations):
        """
        decorator that registers a new implementation and returns the descriptor
        """
        def register_selector_decorator(func):
            assert not isinstance(func, type(self))
            self._add_implementations(implementations, func)
            return self

        return register_selector_decorator

    def __get__(self, instance, *_ignored):
        if instance is None:
            return self
        return _ImplementationBindingMethod(instance=instance, selector=self)



"""

    utility classes for implementation lookup
"""
from collections import Mapping
_NOT_GIVEN = object()


class ImplementationName(object):
    """
    utility class to declare names for implementations
    """
    def __init__(self, name, doc=None):
        self.name = name
        self.documentation = doc

    def __repr__(self):
        return "<Implementation {name}>".format(name=self.name)


class AttributeBasedImplementations(Mapping):

    def __init__(self, holder, attribute_mapping):
        self.holder = holder
        self.attribute_mapping = attribute_mapping

    def __getitem__(self, key):
        attribute_name = self.attribute_mapping[key]
        result = getattr(self.holder, attribute_name, _NOT_GIVEN)
        if result is _NOT_GIVEN:
            raise LookupError('{holder!r} has no attribute {name}'.format(
                holder=self.holder,
                name=attribute_name,
            ))
        return result

    def __iter__(self):
        return iter(self.attribute_mapping)

    def __len__(self):
        return len(self.attribute_mapping)



#!env python
"""
  Sentaku pypi search exampple
  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~

  before running::

    $ pip install selenium  requests sentaku
"""
import argparse
import contextlib
import sentaku
import requests
from selenium.webdriver import Remote
from selenium.webdriver.common.desired_capabilities import DesiredCapabilities

parser = argparse.ArgumentParser()
parser.add_argument('query')
parser.add_argument('--fast', action='store_true')


FastSearch = sentaku.ImplementationName(
    'FastSearch',
    """Used to implement fast search""")

Browser = sentaku.ImplementationName(
    'FastSearch',
    """used to implement all browser actions and browser based slow search""")


class Search(sentaku.Element):
    """sentaku element for really simple pypi searching"""
    search = sentaku.ContextualMethod()

    @search.implemented_for(Browser)
    def search(self, text):
        """do a slow search via the website and return the first match"""
        self.impl.get('https://pypi.python.org/pypi')

        search_div = self.impl.find_element_by_id('search')
        search_term = search_div.find_element_by_id('term')
        search_term.send_keys(text)
        search_div.find_element_by_id('submit').click()
        e = self.impl.find_element_by_css_selector('table.list tr td a')
        return e.get_attribute('href')

    @search.implemented_for(FastSearch)
    def search(self, text):
        """do a sloppy quick "search" via the json index"""

        resp = self.impl.get(
            'https://pypi.python.org/pypi/{text}/json'.format(text=text))
        return resp.json()['info']['package_url']

    open_page = sentaku.ContextualMethod()

    @open_page.implemented_for(Browser)
    def open_page(self, url):
        self.impl.get(url)


def main(search, query):
    """main function that does the search"""
    url = search.search(query)
    print(url)
    search.open_page(url)


def cli_main():
    """cli entrypoitns, sets up everything needed"""
    args = parser.parse_args()
    # open up a browser
    firefox_remote = Remote(
        'http://127.0.0.1:4444/wd/hub',
        DesiredCapabilities.FIREFOX)
    with contextlib.closing(firefox_remote):

        search = Search({
            FastSearch: requests,
            Browser: firefox_remote,
        })

        if args.fast:
            with search.use(FastSearch, Browser):
                main(search, args.query)
        else:
            with search.use(Browser):
                main(search, args.query)


if __name__ == '__main__':
    cli_main()



import warnings
from .context import ImplementationContext
from . import ContextualMethod


class ApplicationImplementation(object):
    """Base class for implementations

    subclasses of this class will be used to hold data necessary
    for one particular implementation

    they are also used to name and refer to implementations for selection

    :arg implementation: object holding an actual implementation
    """
    def __init__(self, implementation):
        self.implementation = implementation


class ApplicationDescription(ImplementationContext):
    def __init__(self, *k, **kw):
        warnings.warn(
            "ApplicationDescription is deprecated, use ImplementationContext",
            category=DeprecationWarning, stacklevel=2)
        super(ApplicationDescription, self).__init__(*k, **kw)


def ImplementationRegistry():
    warnings.warn(
        'Renamed to ContextualMethod',
        category=DeprecationWarning, stacklevel=2)
    return ContextualMethod()





from .utils import (
    ImplementationName,
    AttributeBasedImplementations,
)

from .modeling import Element, Collection
from .context import ContextualMethod, ImplementationContext
from .legacy import ApplicationImplementation, ImplementationRegistry, ApplicationDescription

__all__ = [
    'Element',
    'Collection',

    'ContextualMethod',
    'ImplementationContext',

    'ApplicationImplementation',
    'ImplementationRegistry',
    'ApplicationDescription',

    'ImplementationName',
    'AttributeBasedImplementations',
]



import warnings


class Element(object):
    """Base class for all application elements

    :param parent:
        controlling object of the element which is either a context
        or a surrounding element
    :type parent: :py:class:`Element` or :py:class:`ImplementationContext`
    """
    def __init__(self, parent):
        self.parent = parent

    @property
    def context(self):
        """the context this element belongs to"""
        return self.parent.context

    @property
    def root(self):
        warnings.warn(
            "Element.root got renamed to .context",
            category=DeprecationWarning,
            stacklevel=2,
        )
        return self.context

    @property
    def impl(self):
        """shortcut to get the currently active application implementation"""
        return self.context.impl


class Collection(Element):
    """base class for collections in the application

    :todo: generic helpers for querying
    """



class TodoElement(object):
    """
    Element of a todo list
    """
    def __init__(self, name, completed=False):
        self.name = name
        self.completed = completed


class TodoList(object):
    """a named todolist"""
    def __init__(self, name, elements=()):
        self.name = name
        self.items = list(elements)

    def get_by(self, name):
        """get element by name"""
        for item in self.items:
            if item.name == name:
                return item

    def create_item(self, name):
        """create element by name"""
        assert self.get_by(name) is None
        e = TodoElement(name=name)
        self.items.append(e)
        return e

    def clear_completed(self):
        """
        removes completed elements
        """
        self.items = [i for i in self.items if not i.completed]


class TodoApp(object):
    """
    A Basic Todo List Storage

    """

    def __init__(self):
        #: the todo collections
        self.collections = []

    def __repr__(self):
        return '<TodoApp %r>' % (sorted(x.name for x in self.collections))

    def get_by(self, name):
        """get a todo list by name"""
        for item in self.collections:
            if item.name == name:
                return item

    def create_item(self, name):
        """create a new named todo list"""
        assert self.get_by(name) is None
        e = TodoList(name=name)
        self.collections.append(e)
        return e
